#!/bin/sh
# the next line restarts using the interpreter \
exec wish "$0" "$@"

# moodss: a Modular Object Oriented Dynamic SpreadSheet
# version 1.0, Copyright (C) 1997 Jean-Luc Fontaine
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# mailto:jfontain@mygale.org

set rcsId {$Id: moodss.tcl,v 1.43 1998/01/16 20:16:17 jfontain Exp $}

source getopt.tcl

proc printUsage {exitCode} {
    global argv0

    puts stderr "usage: $argv0 \[-update seconds\] module"
    exit $exitCode
}

if {[catch {set argv [parseCommandLineArguments {-update 1} $argv value]} message]} {
    puts stderr $message
    printUsage 1
}

if {[llength $argv]!=1} {
    printUsage 1
}

lappend auto_path .                                                   ;# search in current directory sub-directories for development
set module [lindex $argv 0]
if {[catch {package require $module}]} {
    puts stderr "$argv0: could not load package \"[lindex $argv 0]\""
    exit 1
}

set pollTimes [set ${module}::data(pollTimes)]
if {[llength $pollTimes]==0} {
    error "module $module poll times list is empty"
}
set pollTime [lindex $pollTimes 0]                                                                                    ;# set default
set pollTimes [lsort -integer $pollTimes]                                                                         ;# sort poll times
catch {set pollTime $value(-update)}                                                      ;# which can be overridden by command line
if {$pollTime<[lindex $pollTimes 0]} {
    puts stderr "$argv0: minimum poll time is [lindex $pollTimes 0] seconds for module $module"
    exit 1
}

if {[catch {package require stooop 3.0}]} {
    source stooop.tcl                                                                     ;# in case stooop package is not installed
}
namespace import stooop::*

if {[catch {package require scwoop 2.0}]} {
    source scwoop.tcl                                                                     ;# in case scwoop package is not installed
    source bindings.tcl
    source widgetip.tcl
    source arrowbut.tcl
    source spinent.tcl
}

if {[catch {package require Tktable 1.8}]} {
    load ./Tktable.so.1.80                                                               ;# in case tkTable package is not installed
}

if {[catch {package require blt 8.0}]} {
    load ./libBLT8.0.so                                                                      ;# in case BLT package is not installed
}

source scroll.tcl
source lifolbl.tcl
source keyslink.tcl
source dialog.tcl
source datatab.tcl
source datagraf.tcl
source help.tcl
source switched.tcl
source drag.tcl
source drop.tcl
source canvhand.tcl

proc minimum {a b} {return [expr {$a<$b?$a:$b}]}
proc maximum {a b} {return [expr {$a>$b?$a:$b}]}

proc updateTitle {} {
    global module pollTime
    wm title . "moodss: $module data (every $pollTime seconds)"                         ;# show module and poll time in window title
}

proc createMenuWidget {parentPath} {
    set menu [menu $parentPath.menu -tearoff 0]

    $menu add cascade -label File -menu [menu $menu.file -tearoff 0] -underline 0
    $menu.file add command -label Exit -command exit -underline 1 -accelerator Alt+X
    bind $parentPath <Alt-x> exit

    $menu add cascade -label Options -menu [menu $menu.options -tearoff 0] -underline 0
    $menu.options add command -label {Poll time...} -command inquirePollTime -underline 0

    $menu add cascade -label Help -menu [menu $menu.help -tearoff 0] -underline 0
    $menu.help add command -label Module... -underline 0 -accelerator F1 -command {
        set text {no help available for this module}
        catch {set text [set ${module}::data(helpText)]}
        simpleTextDialogBox "moodss: help for module $module" $text
    }
    bind $parentPath <F1> help
    $menu.help add command -label Copyright... -underline 0 -command {simpleTextDialogBox {moodss: Copyright} $help::copyright}
    $menu.help add command -label Warranty... -underline 0 -command {simpleTextDialogBox {moodss: Warranty} $help::warranty}
    $menu.help add command -label About... -underline 0 -command aboutDialogBox

    $parentPath configure -menu $menu
}

proc createMessageWidget {parentPath} {
    global messenger

    set messenger [new lifoLabel $parentPath]
    return $widget::($messenger,path)                                                                     ;# return actual tk widget
}

proc createDragAndDropZone {parentPath} {
    global canvas

    set frame [frame $parentPath.drops]
    set icon {
        R0lGODlhIAAgAMIAAP8A/wAAAIKCgv////sUQVFR+zz7NAAAACH5BAEAAAAALAAAAAAgACAAAAOs
        CLrc/jDKqYK4OOudXRhgKI5kKHhlqp7Np74iWwUuDMtLba8oSMCEHwlX8Q2EJSHShFImkUsQkQYl
        RaMDIkB3HGGXBC0XnAz9xFZjKihEe9U2HI3rixYGhfvI/crjQX54BWIGfSmBaAaKJHo3PSCKhX+Q
        iyVTdAOFepKZlo8jBZEjnEwtJY0lolIeea2BNot8O6lyc7a3uLk0HgwBvb85wMIUEb7Ex8jJyssM
        CQA7
    }
    set label [label $frame.graph -image [image create photo -data $icon]]
    pack $label -side left
    new dropSite -path $label -formats DATACELL -command "dataGraph::new \$dragSite::data(DATACELL) $canvas"
    new widgetTip . -path $label -text {graph drop site}
    return $frame
}

proc inquirePollTime {} {
    global pollTimes pollTime table

    set dialog\
        [new dialogBox . -buttons oc -default o -title {moodss: Poll time} -die 0 -x [winfo pointerx .] -y [winfo pointery .]]
    set frame [frame $widget::($dialog,path).frame]
    set minimum [lindex $pollTimes 0]
    set message [message $frame.message\
        -width [winfo screenwidth .] -font {Helvetica -12} -justify center\
        -text "Please enter new poll time\n(greater than $minimum):"\
    ]
    pack $message
    set entry [new spinEntry $frame -width 4 -list $pollTimes -side right]
    spinEntry::set $entry $pollTime
    pack $widget::($entry,path) -anchor e -side left -expand 1 -padx 2       ;# evenly pack entry and label together near the center
    pack [label $frame.label -text seconds] -anchor w -side right -expand 1 -padx 2
    dialogBox::display $dialog $frame
    widget::configure $dialog -command "
        set time \[spinEntry::get $entry\]
        if {\$time<$minimum} {                                                                    ;# check against minimum poll time
            bell
            $message configure -text {Please enter new poll time\n(must be greater than $minimum):}
        } else {                                                                                                ;# new time is valid
            if {\$time!=\$pollTime} {                                             ;# but check that it actually differs from current
                set pollTime \$time
                updateTitle
                refresh                                       ;# update imediately in case poll time was set to a much greater value
            }
            delete $dialog                                                                                     ;# destroy dialog box
        }
    "
    bind $widget::($dialog,path) <Destroy> "catch {delete $entry}"                   ;# delete objects not managed by the dialog box
}

proc refresh {} {
    global updateEvent module messenger pollTime

    catch {after cancel $updateEvent}                                                             ;# eventually cancel current event

    lifoLabel::push $messenger "launching $module data update..."               ;# we do not know when data will actually be updated
    update idletasks
    ${module}::update                                                                       ;# ask module to update its dynamic data
    lifoLabel::pop $messenger
    update idletasks                                                                      ;# restore original cursor at the earliest
    set updateEvent [after [expr {1000*$pollTime}] refresh]                                               ;# convert to milliseconds
}

proc dataTableTitleColumnMotion {path x y} {                    ;# x and y must be set to empty strings when cursor leaves the table
    global tableData messenger module

    if {[catch {scan [$path index @$x,$y] %d,%d row column}]||($row>=0)} {                                      ;# not the title row
        if {[info exists tableData(lastColumn)]} {                                                ;# cursor was inside the title row
            lifoLabel::pop $messenger
            widgetTip::disable $tableData(tip)
            unset tableData(lastColumn)
        }
    } elseif {![info exists tableData(lastColumn)]} {                                            ;# cursor is entering the title row
        lifoLabel::push $messenger [set ${module}::data([set tableData(lastColumn) $column],message)]
        widgetTip::enable $tableData(tip)
        set tableData(lastColumn) $column
    } elseif {$column!=$tableData(lastColumn)} {                                   ;# cursor enters another row within the title row
        lifoLabel::pop $messenger
        lifoLabel::push $messenger [set ${module}::data([set tableData(lastColumn) $column],message)]
    }
}

proc dataTableHelpSetup {path} {                                   ;# setup context sensitive help and widget tip help on data table
    global tableData

    bind $path <Motion> {dataTableTitleColumnMotion %W %x %y}
    bind $path <Leave> {dataTableTitleColumnMotion %W {} {}}
    set tableData(tip) [new widgetTip . -text {click to toggle sort}]         ;# use a custom managed widget tip for help on sorting
}

createMenuWidget .
pack [createMessageWidget .] -side bottom -fill x
updateTitle

set view [new scroll canvas .]
set canvas $composite::($view,scrolled,path)
set width [winfo screenwidth .]
set height [winfo screenheight .]
$canvas configure -background white -width $width -height $height -scrollregion [list 0 0 $width $height]
wm geometry . 400x300

pack [createDragAndDropZone .] -fill x
pack $widget::($view,path) -fill both -expand 1

set table [new dataTable $canvas -data ${module}::data]
new canvasHandles $canvas -path $widget::($table,path)

dataTableHelpSetup $dataTable::($table,tablePath)

refresh                                                                                                ;# initialize refresh process
